

/**
* This file is part of license combination gpl version 3 license and eCos.
* The corresponding license terms are below.
*
* gpl version 3 Licence:
*
* The file were developed during the student thesis "Data Aggregation 
* using TinyIPFIX in Wireless Sensor Networks" of Benjamin Ertl and is 
* included in the dissertation "Secure Data Transmission in Wireless 
* Sensor Networks" by Corinna Schmitt during employment at the Technische
* Universitt Mnchen, Department Computer Science, Chair Network 
* Architectures and Services (Germany).
* Copyright (C) 2013  
* Authors: Benjamin Ertl and Corinna Schmitt (schmitt[at]net.in.tum.de)
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, version 3 of the License
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*
*
*
* eCos Licence: (http://ecos.sourceware.org)
*
* This file is part of eCos, the Embedded Configurable Operating System. 
* Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005, 2006, 
* 2007, 2008, 2009 Free Software Foundation, Inc. 
*
* eCos is free software; you can redistribute it and/or modify it under 
* the terms of the GNU General Public License as published by the Free 
* Software Foundation; either version 3 any later version. 
* 
* eCos is distributed in the hope that it will be useful, but WITHOUT ANY 
* WARRANTY; without even the implied warranty of MERCHANTABILITY or 
* FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
* more details. 
* 
* You should have received a copy of the GNU General Public License along 
* with eCos; if not, write to the Free Software Foundation, Inc., 51 
* Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA. 
*
* As a special exception, if other files instantiate templates or use 
* macros or inline functions from this file, or you compile this file and * link it with other works to produce a work based on this file, this file 
* does not by itself cause the resulting work to be covered by the GNU 
* General Public License. However the source code for this file must still 
* be made available in accordance with section (3) of the GNU General 
* Public License v2. 
* 
* This exception does not invalidate any other reasons why a work based on 
* this file might be covered by the GNU General Public License.
*
*/


package ipfix;
import java.io.*;
import java.util.Calendar;
import java.util.LinkedList;

import xml.MetadataHeap;

public class IPFIXParser {
	
	public static void parse(byte[] packet, String source) throws Exception{
		DataInputStream stream = new DataInputStream(new ByteArrayInputStream(packet));
		int header = stream.readUnsignedShort();
		int availiable = 14;
		//packets with length of 10 bytes need to be passed to the translator, they just match the ipfix version field by accident.
		if(header != 0x000a || packet.length == 10){
			packet = translateHeader(packet, source);
			stream = new DataInputStream(new ByteArrayInputStream(packet));
			availiable = 16;
			stream.skip(2);
		}
		if(stream.available() > availiable){
			int len = stream.readUnsignedShort();
			if(packet.length < len){
				throw new Exception("Package too short to contain the payload");
			}
			byte[] buff = new byte[len];
			
			stream.skip(12);
			int setID = stream.readUnsignedShort();
			if(setID == 2){
				stream.close();
				System.arraycopy(packet, 0, buff, 0, len);
				parseTemplate(buff, source);
			}else{
				stream.close();
				System.arraycopy(packet, 0, buff, 0, len);
				parseData(buff, source);
			}
		}
	}
	
	public static byte[] translateHeader(byte[] packet, String source) throws Exception{
		DataInputStream stream = new DataInputStream(new ByteArrayInputStream(packet));
		//unsigned integers are needed for the bit-shifting that follows.
		int byte1 = stream.readUnsignedByte();
		int byte2 = stream.readUnsignedByte();
		
		stream.close();
		byte[] returner = null;
		//construct the length
		int len = (byte1 & 0x03) + byte2;
		//construct the flags
		int flags = byte1 >> 6;
		//construct the set lookup field
		int set_lookup = (byte1 & 0x3F) >> 2;
		
		if(packet.length != len){
			throw new Exception("invalid short IPFIX Packet, cannot translate header, length is not matching! It is "+packet.length+" but is said to be "+len);
		}
		
		//calculate payload length from flags
		if(flags == 0) len -= 3;
		else if (flags == 3) len -= 5;
		else len -= 4;
		//array that will hold the package with a full IPFIX Header
		returner = new byte[20+len];
		
		//write the version number
		returner[1] = 0x0a;
		
		//write the length
		returner[2] = (byte)(returner.length >> 8);
		returner[3] = (byte)(returner.length & 0xff);
		
		//write system time, for lack of better values use the current time
		long now = System.currentTimeMillis();
		returner[4] = (byte)((now & 0xff000000l) >> 24);
		returner[5] = (byte)((now & 0x00ff0000l) >> 16);
		returner[6] = (byte)((now & 0x0000ff00l) >> 8);
		returner[7] = (byte)((now & 0x000000ffl) >> 0);
		
		//write sequence id, handle extended sequence field
		if(flags == 1 || flags == 3){
			returner[10] = packet[2];
			returner[11] = packet[3];
		}else{
			returner[11] = packet[2];
		}
		
		//get the node ID from the IPv6 address that was given and write it to the observation domain id field
		String obsDomain = "0x"+source.replaceAll("\\[([a-fA-F0-9]*:)*|\\]:[a-fA-F0-9]*", "");
		int obsdomain = Integer.decode(obsDomain);		
		returner[14] = (byte)((obsdomain & 0xff00) >> 8);
		returner[15] = (byte)((obsdomain & 0x00ff) >> 0);
		
		//write the set id
		if(flags == 2 || flags == 3){
			//extended field is present, check if it is used
			if(set_lookup == 0x0F) if(flags == 2) returner[17] = packet[3]; else returner[17] = packet[4];
			else if(set_lookup == 0) if(flags == 2) returner[16] = packet[3]; else returner[16] = packet[4];
			else if(set_lookup == 1) returner[17] = 2;
			else if(set_lookup == 2) returner[16] = 1;
		}else{
			if(set_lookup == 1) returner[17] = 2;
			else if(set_lookup == 2) returner[16] = 1;
		}
		
		//write the set length
		returner[18] = (byte)(((returner.length - 16) & 0xff00) >> 8);
		returner[19] = (byte)((returner.length - 16) & 0xff);
		
		//copy payload
		System.arraycopy(packet, packet.length - len, returner, 20, len);
		
		return returner;
	}
	
	public static void parseTemplate(byte[] packet, String source)throws Exception{
		int validIPFIXHeader = 0x000a;
		int lastvalue = 0;
		Template currentTemplate = null;
		UnsignedDataInputStream stream = new UnsignedDataInputStream(new ByteArrayInputStream(packet));

		if(stream.available() > 23 && (lastvalue = stream.readUnsignedShort()) == validIPFIXHeader){
			if((lastvalue = stream.readUnsignedShort()) < packet.length){
				throw new Exception("The specified length does not match the length of the packet. It should be " + lastvalue + " but it is "+ packet.length+ " Byte");
			}
			
			long export = stream.readUnsignedInt();
			long sequence = stream.readUnsignedInt();
			long observation = stream.readUnsignedInt();
			
			

			while(stream.available() > 0){
				if((lastvalue = stream.readUnsignedShort()) != 2){
					throw new Exception("Template sets must have a Set ID of 2, this one has "+ lastvalue+ "!");
				}
					
				int length = stream.readUnsignedShort();
				
				
				int bytesread = 4;
				while(bytesread < length){
					int templateID = stream.readUnsignedShort();
					bytesread += 2;
					int fieldcount = stream.readUnsignedShort();
					bytesread += 2;
					
					currentTemplate = new Template(templateID,source,observation);
					System.out.printf("+--%s[%d], Template: %d, received %s \n|\n",source,observation,templateID,Calendar.getInstance().getTime().toLocaleString());
					for(int i = 0; i < fieldcount; i++){
						int fieldID = stream.readUnsignedShort();
						bytesread += 2;
						int fieldLength = stream.readUnsignedShort();
						bytesread += 2;
						long enterpriseID = 0;
						if(fieldID > 0x8000){
							enterpriseID = stream.readUnsignedInt();
							bytesread += 4;
						}
						Field currentfield = new Field(fieldID, enterpriseID, fieldLength); 
						//enrich with metadata from XML definition
						MetadataHeap.getInstance().updateField(templateID, source, observation, currentfield);
						currentTemplate.addField(currentfield);	
						System.out.printf("|---- %s[%d] (%d - %d)\n", currentfield.getName(), currentfield.getLength(), currentfield.getEnterpriseID(), currentfield.getID());
					}	
					TemplateHeap.getInstance().putTemplate(currentTemplate);
				}
				System.out.printf("\n");
			}

		}else{
			if (stream.available() <= 23){
				throw new Exception("Package of insufficient length, it must be at least 24 Byte to contain a IPFIX Template");
			}
			if(lastvalue != validIPFIXHeader){
				throw new Exception("This parser is for version " + validIPFIXHeader + " of IPFIX, but the value in the header said "+lastvalue);
			}
		}
	}
	
	public static void parseData(byte[] packet, String source)throws Exception{
		Template currentTemplate = null;
		int validIPFIXHeader = 0x000a;
		int lastvalue = 0;

		UnsignedDataInputStream stream = new UnsignedDataInputStream(new ByteArrayInputStream(packet));

		if(stream.available() > 16 && (lastvalue = stream.readUnsignedShort()) == validIPFIXHeader){
			if((lastvalue = stream.readUnsignedShort()) != packet.length){
				throw new Exception("The specified length does not match the length of the packet. It should be " + lastvalue + " but it is "+ packet.length+ " Byte");
			}

			long export = stream.readUnsignedInt();
			long sequence = stream.readUnsignedInt();			
			long observation = stream.readUnsignedInt();
			LinkedList<Field> fields = new LinkedList<Field>();
						
			while(stream.available() > 0){
				if((lastvalue = stream.readUnsignedShort()) < 256){
					throw new Exception("Datasets must have Set IDs > 255, this one has "+ lastvalue+ "!");
				}
				
				System.out.printf("+--%s[%d], Data: %d received %s\n|\n",source,observation,lastvalue,Calendar.getInstance().getTime().toLocaleString());
				
				currentTemplate = TemplateHeap.getInstance().getTemplate(lastvalue, source, observation);
				if(currentTemplate != null){
					fields = currentTemplate.getFields();
				} else{
					System.out.printf("|---- No template, raw data:\n|---- ");
				}
				int setlength = stream.readUnsignedShort();				
												
				Field currentField;
				int bytesread = 4;
				while(bytesread < setlength){
					if(currentTemplate != null){
					currentField = fields.peekFirst();
					if(sequence <= currentTemplate.getSequence()){
						System.err.println("Got a data set with a obsolete sequence ID, it was ignored.");
						break;
					}
					
					int fieldlen = currentField.getLength();
					if(stream.available() < fieldlen){
						throw new Exception("Field length is longer than the remainder of the packet.");
					}
					bytesread += fieldlen;
					if(currentField.getDatatype().equals("int8")){
						currentField.setValue(stream.readUnsignedByte());
					} else if(currentField.getDatatype().equals("int16")){
						currentField.setValue(stream.readUnsignedShort());
					}else if (currentField.getDatatype().equals("int32")){
						currentField.setValue(stream.readUnsignedInt());
					}else if(currentField.getDatatype().equals("float")){
						currentField.setValue(stream.readFloat());
					}else{
						byte[] buffer = new byte[fieldlen];
						stream.readFully(buffer);
						currentField.setValue(buffer);
					}
					//rotate the list one element further;
					fields.addLast(fields.removeFirst());
					System.out.printf("|---- %s[%d] (%d - %d):%s\n", currentField.getName(), currentField.getLength(), currentField.getEnterpriseID(), currentField.getID(), currentField.getValue().toString());
					}else{
						//Template unknown. Just print out the raw data, nothing get's updated back at authone.
						System.out.printf("%02X ",stream.readUnsignedByte() );
						bytesread ++;
					}
				}
				System.out.printf("\n");
				if(currentTemplate == null) System.out.printf("\n");
			}

		}else{
			if (stream.available() <= 16){
				while(stream.available() > 0){
					System.out.printf("%02X ",stream.readUnsignedByte() );
				}
				throw new Exception("Package of insufficient length, it must be at least 17 Byte to contain a IPFIX Dataset. \n");
				
			}
			if(lastvalue != validIPFIXHeader){
				throw new Exception("This parser is for version " + validIPFIXHeader + " of IPFIX, but the value in the header said "+lastvalue);
			}
		}
	}	
}
